import os

import numpy as np
import pytest

from qililab.qprogram import QProgram
from qililab.qprogram.blocks import Block
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.waveforms import IQPair, Square
from qililab.yaml import yaml


class TestCalibration:
    """Unit tests checking the Calibration attributes and methods."""

    def test_init(self):
        """Test init method"""
        calibration = Calibration()

        assert isinstance(calibration, Calibration)
        assert isinstance(calibration.waveforms, dict)
        assert isinstance(calibration.weights, dict)
        assert isinstance(calibration.blocks, dict)
        assert calibration.crosstalk_matrix is None
        assert len(calibration.waveforms) == 0
        assert len(calibration.weights) == 0
        assert len(calibration.blocks) == 0

    def test_add_waveform_method(self):
        """Test add_waveform method"""
        xpi = Square(1.0, 100)
        xpi2 = Square(1.0, 50)
        readout = Square(1.0, 2000)

        calibration = Calibration()
        calibration.add_waveform(bus="drive_bus", name="Xpi", waveform=xpi)

        assert len(calibration.waveforms) == 1
        assert "drive_bus" in calibration.waveforms

        assert isinstance(calibration.waveforms["drive_bus"], dict)
        assert len(calibration.waveforms["drive_bus"]) == 1
        assert "Xpi" in calibration.waveforms["drive_bus"]
        assert calibration.waveforms["drive_bus"]["Xpi"] == xpi

        calibration.add_waveform(bus="drive_bus", name="Xpi2", waveform=xpi2)
        assert len(calibration.waveforms["drive_bus"]) == 2
        assert "Xpi2" in calibration.waveforms["drive_bus"]
        assert calibration.waveforms["drive_bus"]["Xpi2"] == xpi2

        calibration.add_waveform(bus="readout_bus", name="readout", waveform=readout)
        assert len(calibration.waveforms) == 2
        assert isinstance(calibration.waveforms["readout_bus"], dict)
        assert len(calibration.waveforms["readout_bus"]) == 1
        assert "readout" in calibration.waveforms["readout_bus"]
        assert calibration.waveforms["readout_bus"]["readout"] == readout

    def test_add_weights_method(self):
        """Test add_weights method"""
        weights = IQPair(Square(1.0, 2000), Square(1.0, 2000))

        calibration = Calibration()
        calibration.add_weights(bus="readout_bus", name="optimal_weights", weights=weights)

        assert len(calibration.weights) == 1
        assert "readout_bus" in calibration.weights

        assert isinstance(calibration.weights["readout_bus"], dict)
        assert len(calibration.weights["readout_bus"]) == 1
        assert "optimal_weights" in calibration.weights["readout_bus"]
        assert calibration.weights["readout_bus"]["optimal_weights"] == weights

    def test_add_block_method(self):
        """Test add_weights method"""
        qp = QProgram()
        qp.play(bus="flux_x", waveform=Square(1.0, 100))
        qp.play(bus="flux_y", waveform=Square(1.0, 100))

        calibration = Calibration()
        calibration.add_block(name="flux_block", block=qp.body)

        assert len(calibration.blocks) == 1
        assert "flux_block" in calibration.blocks
        assert isinstance(calibration.blocks["flux_block"], Block)

    def test_has_waveform_method(self):
        """Test has_operation method"""
        xpi = Square(1.0, 100)
        xpi2 = Square(1.0, 50)

        calibration = Calibration()
        calibration.add_waveform(bus="drive_bus", name="Xpi", waveform=xpi)
        calibration.add_waveform(bus="drive_bus", name="Xpi2", waveform=xpi2)

        assert calibration.has_waveform(bus="drive_bus", name="Xpi") is True
        assert calibration.has_waveform(bus="drive_bus", name="Xpi2") is True
        assert calibration.has_waveform(bus="drive_bus", name="non_existant") is False
        assert calibration.has_waveform(bus="non_existant", name="Xpi") is False

    def test_has_weights_method(self):
        """Test add_operation method"""
        weights = IQPair(Square(1.0, 2000), Square(1.0, 2000))

        calibration = Calibration()
        calibration.add_weights(bus="readout_bus", name="optimal_weights", weights=weights)

        assert len(calibration.weights) == 1
        assert "readout_bus" in calibration.weights

        assert calibration.has_weights(bus="readout_bus", name="optimal_weights") is True
        assert calibration.has_weights(bus="readout_bus", name="non_existant_weights") is False
        assert calibration.has_weights(bus="non_existant", name="non_existant_weights") is False

    def test_has_block_method(self):
        """Test add_weights method"""
        qp = QProgram()
        qp.play(bus="flux_x", waveform=Square(1.0, 100))
        qp.play(bus="flux_y", waveform=Square(1.0, 100))

        calibration = Calibration()
        calibration.add_block(name="flux_block", block=qp.body)

        assert calibration.has_block("flux_block")

    def test_get_waveform_method(self):
        """Test get_operation method"""
        xpi = Square(1.0, 100)
        xpi2 = Square(1.0, 50)

        calibration = Calibration()
        calibration.add_waveform(bus="drive_bus", name="Xpi", waveform=xpi)
        calibration.add_waveform(bus="drive_bus", name="Xpi2", waveform=xpi2)

        retrieved_xpi = calibration.get_waveform(bus="drive_bus", name="Xpi")
        assert xpi == retrieved_xpi

        retrieved_xpi2 = calibration.get_waveform(bus="drive_bus", name="Xpi2")
        assert xpi2 == retrieved_xpi2

        with pytest.raises(KeyError):
            _ = calibration.get_waveform(bus="drive_bus", name="non_existant")

        with pytest.raises(KeyError):
            _ = calibration.get_waveform(bus="non_existant", name="Xpi")

    def test_get_weights_method(self):
        """Test add_operation method"""
        weights = IQPair(Square(1.0, 2000), Square(1.0, 2000))

        calibration = Calibration()
        calibration.add_weights(bus="readout_bus", name="optimal_weights", weights=weights)

        retrieved_weights = calibration.get_weights(bus="readout_bus", name="optimal_weights")
        assert retrieved_weights == weights

        with pytest.raises(KeyError):
            _ = calibration.get_weights(bus="readout_bus", name="non_existant_weights")

        with pytest.raises(KeyError):
            _ = calibration.get_weights(bus="non_existant_bus", name="optimal_weights")

    def test_get_block_method(self):
        """Test add_weights method"""
        qp = QProgram()
        qp.play(bus="flux_x", waveform=Square(1.0, 100))
        qp.play(bus="flux_y", waveform=Square(1.0, 100))

        calibration = Calibration()
        calibration.add_block(name="flux_block", block=qp.body)

        retrieved_block = calibration.get_block(name="flux_block")
        assert retrieved_block == qp.body

        with pytest.raises(KeyError):
            _ = calibration.get_block(name="non_existant_block")

    def test_adding_crosstalk_matrix(self):
        """Test adding a crosstalk matrix to the calibration object"""
        buses = {
            "flux_0": {"flux_0": 1.47046905, "flux_1": 0.12276261},
            "flux_1": {"flux_0": -0.55322207, "flux_1": 1.58247856},
        }
        crosstalk_matrix = CrosstalkMatrix().from_buses(buses)
        calibration = Calibration()
        calibration.crosstalk_matrix = crosstalk_matrix

        assert calibration.crosstalk_matrix["flux_0"]["flux_0"] == crosstalk_matrix["flux_0"]["flux_0"]
        assert calibration.crosstalk_matrix["flux_0"]["flux_1"] == crosstalk_matrix["flux_0"]["flux_1"]
        assert calibration.crosstalk_matrix["flux_1"]["flux_0"] == crosstalk_matrix["flux_1"]["flux_0"]
        assert calibration.crosstalk_matrix["flux_1"]["flux_1"] == crosstalk_matrix["flux_1"]["flux_1"]

    def test_dump_load_methods(self):
        """Test dump and load methods"""
        qp = QProgram()
        qp.play(bus="flux_x", waveform=Square(1.0, 100))
        qp.play(bus="flux_y", waveform=Square(1.0, 100))

        buses = {
            "flux_0": {"flux_0": 1.47046905, "flux_1": 0.12276261},
            "flux_1": {"flux_0": -0.55322207, "flux_1": 1.58247856},
        }
        crosstalk_matrix = CrosstalkMatrix().from_buses(buses)
        calibration = Calibration()
        calibration.add_waveform(bus="drive_bus", name="Xpi", waveform=Square(1.0, 100))
        calibration.add_waveform(bus="drive_bus", name="Xpi2", waveform=Square(1.0, 50))
        calibration.add_waveform(bus="readout_bus", name="readout", waveform=Square(1.0, 2000))
        calibration.add_weights(
            bus="readout_bus", name="optimal_weights", weights=IQPair(Square(1.0, 2000), Square(1.0, 2000))
        )
        calibration.add_block(name="flux_block", block=qp.body)
        calibration.crosstalk_matrix = crosstalk_matrix

        calibration.save_to(file="calibration.yml")
        loaded_calibration = Calibration.load_from(file="calibration.yml")

        assert isinstance(loaded_calibration, Calibration)

        xpi = loaded_calibration.get_waveform(bus="drive_bus", name="Xpi")
        xpi2 = loaded_calibration.get_waveform(bus="drive_bus", name="Xpi2")
        readout = loaded_calibration.get_waveform(bus="readout_bus", name="readout")
        weights = calibration.get_weights(bus="readout_bus", name="optimal_weights")
        block = calibration.get_block(name="flux_block")

        assert isinstance(xpi, Square)
        assert xpi.amplitude == 1.0
        assert xpi.duration == 100

        assert isinstance(xpi2, Square)
        assert xpi2.amplitude == 1.0
        assert xpi2.duration == 50

        assert isinstance(readout, Square)
        assert readout.amplitude == 1.0
        assert readout.duration == 2000

        assert isinstance(weights, IQPair)
        assert isinstance(weights.I, Square)
        assert weights.I.amplitude == 1.0
        assert weights.I.duration == 2000
        assert isinstance(weights.Q, Square)
        assert weights.Q.amplitude == 1.0
        assert weights.Q.duration == 2000

        assert isinstance(block, Block)
        assert len(block.elements) == 2

        assert calibration.crosstalk_matrix["flux_0"]["flux_0"] == crosstalk_matrix["flux_0"]["flux_0"]
        assert calibration.crosstalk_matrix["flux_0"]["flux_1"] == crosstalk_matrix["flux_0"]["flux_1"]
        assert calibration.crosstalk_matrix["flux_1"]["flux_0"] == crosstalk_matrix["flux_1"]["flux_0"]
        assert calibration.crosstalk_matrix["flux_1"]["flux_1"] == crosstalk_matrix["flux_1"]["flux_1"]

        os.remove(path="calibration.yml")

        # Test that loading a different yaml produces an error
        square = Square(1.0, 100)

        with open(file="calibration.yml", mode="w", encoding="utf-8") as stream:
            yaml.dump(data=square, stream=stream)

        with pytest.raises(TypeError):
            _ = Calibration.load_from(file="calibration.yml")

        os.remove(path="calibration.yml")

    def test_load_crosstalk_from_array(self):
        """Test dump and load methods"""
        matrix = [[1.47046905, 0.12276261], [-0.55322207, 1.58247856]]
        flux_offsets = [0.0, 0.1]
        bus_list = ["flux_0", "flux_1"]
        calibration = Calibration()
        calibration.crosstalk_matrix = CrosstalkMatrix()
        calibration.crosstalk_matrix.matrix = matrix
        calibration.crosstalk_matrix.flux_offsets = flux_offsets
        calibration.crosstalk_matrix.bus_list = bus_list

        calibration.save_to(file="calibration.yml")
        loaded_calibration = Calibration.load_from(file="calibration.yml")

        crosstalk_dict = CrosstalkMatrix.from_array(
            buses=["flux_0", "flux_1"], matrix_array=np.array([[1.47046905, 0.12276261], [-0.55322207, 1.58247856]])
        )
        crosstalk_dict.set_offset({"flux_0": 0.0, "flux_1": 0.1})

        assert loaded_calibration.crosstalk_matrix.matrix.matrix == crosstalk_dict.matrix
        assert loaded_calibration.crosstalk_matrix.flux_offsets == crosstalk_dict.flux_offsets
        assert loaded_calibration.crosstalk_matrix.bus_list == bus_list

        os.remove(path="calibration.yml")

        # Test that having an empty bus_list causes an error
        error_string = "Bus list is empty or has the wrong dimensions"

        matrix = [[1.47046905, 0.12276261], [-0.55322207, 1.58247856]]
        flux_offsets = {"flux_0": 0.0, "flux_1": 0.1}
        bus_list = []
        calibration = Calibration()
        calibration.crosstalk_matrix = CrosstalkMatrix()
        calibration.crosstalk_matrix.matrix = matrix
        calibration.crosstalk_matrix.flux_offsets = flux_offsets
        calibration.crosstalk_matrix.bus_list = bus_list

        calibration.save_to(file="calibration.yml")
        with pytest.raises(ValueError, match=error_string):
            _ = Calibration.load_from(file="calibration.yml")

        os.remove(path="calibration.yml")

        matrix = crosstalk_dict.matrix
        flux_offsets = [0.0, 0.1]
        bus_list = []
        calibration = Calibration()
        calibration.crosstalk_matrix = CrosstalkMatrix()
        calibration.crosstalk_matrix.matrix = matrix
        calibration.crosstalk_matrix.flux_offsets = flux_offsets
        calibration.crosstalk_matrix.bus_list = bus_list

        calibration.save_to(file="calibration.yml")
        with pytest.raises(ValueError, match=error_string):
            _ = Calibration.load_from(file="calibration.yml")

        os.remove(path="calibration.yml")



"""Data to use alongside the test suite."""

from typing import Any

from qililab.constants import CONNECTION, INSTRUMENTCONTROLLER, PLATFORM, RUNCARD, AWGTypes
from qililab.typings.enums import (
    AcquireTriggerMode,
    ConnectionName,
    InstrumentControllerName,
    InstrumentName,
    IntegrationMode,
    Parameter,
)


class Galadriel:
    """Test data of the galadriel platform."""

    name = "galadriel"

    digital_compilation_settings: dict[str, Any] = {
        PLATFORM.MINIMUM_CLOCK_TIME: 4,
        PLATFORM.DELAY_BEFORE_READOUT: 0,
        "topology": [[0, 2], [1, 2], [2, 3], [2, 4]],
        "gates": {
            "M(0)": [
                {
                    "bus": "feedline_input_output_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 2000,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "M(1)": [
                {
                    "bus": "feedline_input_output_bus_1",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 2000,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "M(2)": [
                {
                    "bus": "feedline_input_output_bus_2",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 2000,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "I(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 100,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "Drag(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 50,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "Drag(1)": [
                {
                    "bus": "drive_line_q1_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 50,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "X(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 50,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "X(1)": [
                {
                    "bus": "drive_line_q1_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 0,
                        "duration": 50,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "Y(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "Y(1)": [
                {
                    "bus": "drive_line_q1_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "RY(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "RX(0)": [
                {
                    "bus": "drive_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "drag", "num_sigmas": 4, "drag_coefficient": 0},
                    },
                }
            ],
            "CZ(0,1)": [
                {
                    "bus": "flux_line_q1_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "rectangular"},
                        "options": {"q0_phase_correction": 0.1, "q1_phase_correction": 0.2},
                    },
                }
            ],
            "CZ(0, 2)": [
                {
                    "bus": "flux_line_q0_bus",
                    "wait_time": 0,
                    "pulse": {
                        "amplitude": 1.0,
                        "phase": 1.5707963267948966,
                        "duration": 20,
                        "shape": {"name": "rectangular"},
                        "options": {"q0_phase_correction": 0.1, "q2_phase_correction": 0.2},
                    },
                }
            ],
        },
        "buses": {
            "drive_line_q0_bus": {
                "line": "drive",
                "qubits": [0],
                "distortions": [
                    {"name": "lfilter", "a": [1.0, 0.0, 1.0], "auto_norm": True, "b": [0.5, 0.5], "norm_factor": 1.0}
                ],
            },
            "drive_line_q1_bus": {"line": "drive", "qubits": [1]},
            "drive_line_q2_bus": {"line": "drive", "qubits": [1]},
            "feedline_input_output_bus": {
                "line": "readout",
                "qubits": [0],
                "delay": 0,
                "distortions": [],
            },
            "feedline_input_output_bus_1": {
                "line": "readout",
                "qubits": [1],
                "delay": 0,
                "distortions": [],
            },
            "feedline_input_output_bus_2": {
                "line": "readout",
                "qubits": [2],
                "delay": 0,
                "distortions": [],
            },
            "flux_line_q0_bus": {"line": "flux", "qubits": [0]},
        },
    }

    analog_compilation_settings: dict[str, Any] = {
        "flux_control_topology": [
            {"flux": "phix_q0", "bus": "flux_line_q0_bus"},
            {"flux": "phiz_q0", "bus": "flux_line_q0_bus"},
            {"flux": "phix_q1", "bus": "drive_line_q0_bus"},
            {"flux": "phiz_q1", "bus": "drive_line_q0_bus"},
            {"flux": "phix_c0_1", "bus": "flux_line_q0_bus"},
            {"flux": "phiz_c0_1", "bus": "flux_line_q0_bus"},
        ]
    }

    qblox_qcm_0: dict[str, Any] = {
        "name": InstrumentName.QBLOX_QCM.value,
        "alias": InstrumentName.QBLOX_QCM.value,
        AWGTypes.OUT_OFFSETS: [0, 0, 0.7, 0.8],
        AWGTypes.AWG_SEQUENCERS: [
            {
                "identifier": 0,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: True,
            },
            {
                "identifier": 1,
                "outputs": [0],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
            {
                "identifier": 2,
                "outputs": [1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
            {
                "identifier": 3,
                "outputs": [2],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
            {
                "identifier": 4,
                "outputs": [3],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
        ],
    }

    qblox_qcm_rf_0: dict[str, Any] = {
        "name": InstrumentName.QCMRF.value,
        "alias": InstrumentName.QCMRF.value,
        "out0_lo_freq": 3.7e9,
        "out0_lo_en": True,
        "out0_att": 10,
        "out0_offset_path0": 0.2,
        "out0_offset_path1": 0.07,
        "out1_lo_freq": 3.9e9,
        "out1_lo_en": True,
        "out1_att": 6,
        "out1_offset_path0": 0.1,
        "out1_offset_path1": 0.6,
        "awg_sequencers": [
            {
                "identifier": 0,
                "outputs": [0],
                "intermediate_frequency": 20000000,
                "gain_i": 0.001,
                "gain_q": 0.02,
                "gain_imbalance": 1,
                "phase_imbalance": 0,
                "offset_i": 0,
                "offset_q": 0,
                "hardware_modulation": True,
            },
            {
                "identifier": 1,
                "outputs": [1],
                "intermediate_frequency": 20000000,
                "gain_i": 0.001,
                "gain_q": 0.02,
                "gain_imbalance": 1,
                "phase_imbalance": 0,
                "offset_i": 0,
                "offset_q": 0,
                "hardware_modulation": True,
            },
        ],
    }

    qblox_qrm_rf_0: dict[str, Any] = {
        "name": InstrumentName.QRMRF.value,
        "alias": InstrumentName.QRMRF.value,
        "out0_in0_lo_freq": 3e9,
        "out0_in0_lo_en": True,
        "out0_att": 34,
        "in0_att": 28,
        "out0_offset_path0": 0.123,
        "out0_offset_path1": 1.234,
        "awg_sequencers": [
            {
                "identifier": 0,
                "outputs": [0],
                "threshold": 0.5,
                "threshold_rotation": 45.0,
                "intermediate_frequency": 20000000,
                "gain_i": 0.001,
                "gain_q": 0.02,
                "gain_imbalance": 1,
                "phase_imbalance": 0,
                "offset_i": 0,
                "offset_q": 0,
                "hardware_modulation": True,
                "scope_acquire_trigger_mode": "sequencer",
                "scope_hardware_averaging": True,
                "sampling_rate": 1000000000,
                "integration_length": 8000,
                "integration_mode": "ssb",
                "sequence_timeout": 1,
                "acquisition_timeout": 1,
                "hardware_demodulation": True,
                "scope_store_enabled": True,
                "time_of_flight": 40,
            }
        ],
    }

    qblox_qrm_0: dict[str, Any] = {
        "name": InstrumentName.QBLOX_QRM.value,
        "alias": f"{InstrumentName.QBLOX_QRM.value}_0",
        AWGTypes.OUT_OFFSETS: [0.123, 1.23],
        AWGTypes.AWG_SEQUENCERS: [
            {
                "identifier": 0,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_123,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: True,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
            {
                "identifier": 1,
                "outputs": [0, 1],
                Parameter.IF.value: 200_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_000,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: False,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
        ],
    }

    qblox_qrm_1: dict[str, Any] = {
        "name": InstrumentName.QBLOX_QRM.value,
        "alias": f"{InstrumentName.QBLOX_QRM.value}_1",
        AWGTypes.OUT_OFFSETS: [0.123, 1.23],
        AWGTypes.AWG_SEQUENCERS: [
            {
                "identifier": 0,
                "outputs": [0],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_123,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: True,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
        ],
    }

    rohde_schwarz_controller_0: dict[str, Any] = {
        "name": "rohde_schwarz",
        "alias": "rohde_schwarz_controller_0",
        Parameter.REFERENCE_CLOCK.value: "EXT",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.10",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rs_0",
                "slot_id": 0,
            }
        ],
        INSTRUMENTCONTROLLER.RESET: True,
    }

    rohde_schwarz_0: dict[str, Any] = {
        "name": "rohde_schwarz",
        "alias": "rs_0",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 7.24730e09,
        Parameter.RF_ON.value: True,
    }

    rohde_schwarz_controller_1: dict[str, Any] = {
        "name": "rohde_schwarz",
        "alias": "rohde_schwarz_controller_1",
        Parameter.REFERENCE_CLOCK.value: "EXT",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.7",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rs_1",
                "slot_id": 0,
            }
        ],
        INSTRUMENTCONTROLLER.RESET: True,
    }

    rohde_schwarz_1: dict[str, Any] = {
        "name": InstrumentName.ROHDE_SCHWARZ.value,
        "alias": "rs_1",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 3.351e09,
        Parameter.RF_ON.value: True,
    }

    attenuator_controller_0: dict[str, Any] = {
        "name": InstrumentControllerName.MINI_CIRCUITS.value,
        "alias": "attenuator_controller_0",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.222",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "attenuator",
                "slot_id": 0,
            }
        ],
        INSTRUMENTCONTROLLER.RESET: True,
    }

    attenuator: dict[str, Any] = {
        "name": InstrumentName.MINI_CIRCUITS.value,
        "alias": "attenuator",
        Parameter.ATTENUATION.value: 30,
    }

    keithley_2600_controller_0: dict[str, Any] = {
        "name": InstrumentControllerName.KEITHLEY2600.value,
        "alias": "keithley_2600_controller_0",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.112",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": InstrumentName.KEITHLEY2600.value,
                "slot_id": 0,
            }
        ],
        INSTRUMENTCONTROLLER.RESET: True,
    }

    keithley_2600: dict[str, Any] = {
        "name": InstrumentName.KEITHLEY2600.value,
        "alias": InstrumentControllerName.KEITHLEY2600.value,
        Parameter.MAX_CURRENT.value: 0.1,
        Parameter.MAX_VOLTAGE.value: 20.0,
    }

    instruments: list[dict] = [
        qblox_qcm_0,
        qblox_qrm_0,
        qblox_qrm_1,
        qblox_qcm_rf_0,
        qblox_qrm_rf_0,
        rohde_schwarz_0,
        rohde_schwarz_1,
        attenuator,
        keithley_2600,
    ]

    instrument_controllers: list[dict] = [
        rohde_schwarz_controller_0,
        rohde_schwarz_controller_1,
        attenuator_controller_0,
        keithley_2600_controller_0,
    ]

    buses: list[dict[str, Any]] = [
        {
            RUNCARD.ALIAS: "drive_line_q0_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, "rs_0"],
            RUNCARD.CHANNELS: [0, None],
            "distortions": [
                {"name": "lfilter", "a": [1.0, 0.0, 1.0], "auto_norm": True, "b": [0.5, 0.5], "norm_factor": 1.0}
            ],
        },
        {RUNCARD.ALIAS: "drive_line_q1_bus", RUNCARD.INSTRUMENTS: [InstrumentName.QCMRF.value], RUNCARD.CHANNELS: [0]},
        {RUNCARD.ALIAS: "drive_line_q2_bus", RUNCARD.INSTRUMENTS: [InstrumentName.QCMRF.value], RUNCARD.CHANNELS: [1]},
        {
            "alias": "feedline_input_output_bus",
            RUNCARD.INSTRUMENTS: [f"{InstrumentName.QBLOX_QRM.value}_0", "rs_1"],
            RUNCARD.CHANNELS: [0, None],
        },
        {
            "alias": "feedline_input_output_bus_1",
            RUNCARD.INSTRUMENTS: [f"{InstrumentName.QRMRF.value}"],
            RUNCARD.CHANNELS: [0],
        },
        {
            "alias": "feedline_input_output_bus_2",
            RUNCARD.INSTRUMENTS: [f"{InstrumentName.QBLOX_QRM.value}_1"],
            RUNCARD.CHANNELS: [0],
        },
        {
            RUNCARD.ALIAS: "flux_line_q0_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, "rs_0"],
            RUNCARD.CHANNELS: [1, None],
        },
        {
            RUNCARD.ALIAS: "flux_line_q1_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, "rs_0"],
            RUNCARD.CHANNELS: [2, None],
        },
        {
            RUNCARD.ALIAS: "flux_line_q2_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, "rs_0"],
            RUNCARD.CHANNELS: [3, None],
        },
        {
            RUNCARD.ALIAS: "flux_line_q3_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, "rs_0"],
            RUNCARD.CHANNELS: [4, None],
        },
        {
            RUNCARD.ALIAS: "flux_line_too_many_instr",
            RUNCARD.INSTRUMENTS: [InstrumentName.QBLOX_QCM.value, InstrumentName.QBLOX_QCM.value],
            RUNCARD.CHANNELS: [1, 4],
        },
    ]

    runcard: dict[str, Any] = {
        RUNCARD.NAME: name,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
        RUNCARD.BUSES: buses,
        RUNCARD.DIGITAL: digital_compilation_settings,
        RUNCARD.ANALOG: analog_compilation_settings,
    }


class SauronYokogawa:
    """Test data of the sauron with yokogawa platform."""

    name = "sauron_yokogawa"

    yokogawa_gs200_current = {
        RUNCARD.NAME: InstrumentName.YOKOGAWA_GS200,
        RUNCARD.ALIAS: "yokogawa_current",
        Parameter.SOURCE_MODE.value: "current",
        Parameter.CURRENT.value: [0.5],
        Parameter.VOLTAGE.value: [0.0],
        Parameter.SPAN.value: ["200mA"],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        "dacs": [0],
    }

    yokogawa_gs200_voltage = {
        RUNCARD.NAME: InstrumentName.YOKOGAWA_GS200,
        RUNCARD.ALIAS: "yokogawa_voltage",
        Parameter.SOURCE_MODE.value: "voltage",
        Parameter.CURRENT.value: [0.0],
        Parameter.VOLTAGE.value: [0.5],
        Parameter.SPAN.value: ["100mV"],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        "dacs": [0],
    }

    rohde_schwarz: dict[str, Any] = {
        "name": InstrumentName.ROHDE_SCHWARZ.value,
        "alias": "rohde_schwarz",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 7.24730e09,
        Parameter.RF_ON.value: True,
    }

    yokogawa_gs200_current_controller = {
        RUNCARD.NAME: InstrumentControllerName.YOKOGAWA_GS200,
        RUNCARD.ALIAS: "yokogawa_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.15",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "yokogawa_current",
                "slot_id": 0,
            }
        ],
    }

    yokogawa_gs200_voltage_controller = {
        RUNCARD.NAME: InstrumentControllerName.YOKOGAWA_GS200,
        RUNCARD.ALIAS: "yokogawa_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.15",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "yokogawa_voltage",
                "slot_id": 0,
            }
        ],
    }

    yokogawa_gs200_controller_wrong_module = {
        RUNCARD.NAME: InstrumentControllerName.YOKOGAWA_GS200,
        RUNCARD.ALIAS: "yokogawa_controller_wrong_module",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.15",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rohde_schwarz",
                "slot_id": 0,
            }
        ],
    }

    instruments = [yokogawa_gs200_current, yokogawa_gs200_voltage, rohde_schwarz]
    instrument_controllers = [
        yokogawa_gs200_current_controller,
        yokogawa_gs200_voltage_controller,
        yokogawa_gs200_controller_wrong_module,
    ]

    buses: list[dict[str, Any]] = [
        {RUNCARD.ALIAS: "yokogawa_gs200_current_bus", RUNCARD.INSTRUMENTS: ["yokogawa_current"], RUNCARD.CHANNELS: [0]},
        {RUNCARD.ALIAS: "yokogawa_gs200_voltage_bus", RUNCARD.INSTRUMENTS: ["yokogawa_voltage"], RUNCARD.CHANNELS: [0]},
    ]

    runcard = {
        RUNCARD.NAME: name,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
        RUNCARD.BUSES: buses,
    }


class SauronSpiRack:
    """Test data of the sauron with spi rack platform."""

    name = "sauron_spi_rack"

    spi_rack = {
        RUNCARD.NAME: InstrumentName.QBLOX_S4G,
        RUNCARD.ALIAS: "S4g_1",
        Parameter.CURRENT.value: [0.0],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        Parameter.SPAN.value: ["range_min_bi"],
        "dacs": [1],
    }

    rohde_schwarz: dict[str, Any] = {
        "name": InstrumentName.ROHDE_SCHWARZ.value,
        "alias": "rohde_schwarz",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 7.24730e09,
        Parameter.RF_ON.value: True,
    }

    spi_rack_controller_usb = {
        RUNCARD.NAME: InstrumentControllerName.QBLOX_SPIRACK,
        RUNCARD.ALIAS: "spi_controller_usb",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.USB.value,
            CONNECTION.ADDRESS: "ttyUSB0",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "S4g_1",
                "slot_id": 1,
            }
        ],
    }

    spi_rack_controller_wrong_module = {
        RUNCARD.NAME: InstrumentControllerName.QBLOX_SPIRACK,
        RUNCARD.ALIAS: "spi_controller_wrong_module",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.USB.value,
            CONNECTION.ADDRESS: "ttyUSB0",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rohde_schwarz",
                "slot_id": 1,
            }
        ],
    }

    instruments = [spi_rack, rohde_schwarz]
    instrument_controllers = [
        spi_rack_controller_usb,
        spi_rack_controller_wrong_module,
    ]

    buses: list[dict[str, Any]] = [{RUNCARD.ALIAS: "spi_bus", RUNCARD.INSTRUMENTS: ["S4g_1"], RUNCARD.CHANNELS: [1]}]

    runcard = {
        RUNCARD.NAME: name,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
        RUNCARD.BUSES: buses,
    }


class SauronQDevil:
    """Test data of the sauron with yokogawa platform."""

    name = "sauron_qdevil"

    qdevil_qdac2 = {
        RUNCARD.NAME: InstrumentName.QDEVIL_QDAC2,
        RUNCARD.ALIAS: "qdac",
        Parameter.VOLTAGE.value: [0.0],
        Parameter.SPAN.value: ["low"],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        Parameter.LOW_PASS_FILTER.value: ["dc"],
        "dacs": [1],
    }

    rohde_schwarz: dict[str, Any] = {
        "name": InstrumentName.ROHDE_SCHWARZ.value,
        "alias": "rohde_schwarz",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 7.24730e09,
        Parameter.RF_ON.value: True,
    }

    qdevil_qdac2_controller_ip = {
        RUNCARD.NAME: InstrumentControllerName.QDEVIL_QDAC2,
        RUNCARD.ALIAS: "qdac_controller_ip",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.15",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qdac",
                "slot_id": 0,
            }
        ],
    }

    qdevil_qdac2_controller_usb = {
        RUNCARD.NAME: InstrumentControllerName.QDEVIL_QDAC2,
        RUNCARD.ALIAS: "qdac_controller_usb",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.USB.value,
            CONNECTION.ADDRESS: "ttyUSB0",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qdac",
                "slot_id": 0,
            }
        ],
    }

    qdevil_qdac2_controller_wrong_module = {
        RUNCARD.NAME: InstrumentControllerName.QDEVIL_QDAC2,
        RUNCARD.ALIAS: "qdac_controller_wrong_module",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.15",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rohde_schwarz",
                "slot_id": 0,
            }
        ],
    }

    instruments = [qdevil_qdac2, rohde_schwarz]
    instrument_controllers = [
        qdevil_qdac2_controller_ip,
        qdevil_qdac2_controller_usb,
        qdevil_qdac2_controller_wrong_module,
    ]

    buses: list[dict[str, Any]] = [{RUNCARD.ALIAS: "qdac_bus", RUNCARD.INSTRUMENTS: ["qdac"], RUNCARD.CHANNELS: [1]}]

    runcard = {
        RUNCARD.NAME: name,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
        RUNCARD.BUSES: buses,
    }


class SauronVNA:
    """Test data of the sauron platform."""

    name = "sauron_vna"

    keysight_e5080b_controller: dict[str, Any] = {
        "name": InstrumentControllerName.KEYSIGHT_E5080B,
        "alias": InstrumentControllerName.KEYSIGHT_E5080B.value,
        Parameter.TIMEOUT.value: 10000,
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.254",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": InstrumentName.KEYSIGHT_E5080B.value,
                "slot_id": 0,
            }
        ],
    }

    keysight_e5080b: dict[str, Any] = {
        "name": InstrumentName.KEYSIGHT_E5080B,
        "alias": InstrumentName.KEYSIGHT_E5080B.value,
        Parameter.POWER.value: -60.0,
    }

    agilent_e5071b_controller: dict[str, Any] = {
        "name": InstrumentControllerName.AGILENT_E5071B,
        "alias": InstrumentControllerName.AGILENT_E5071B.value,
        Parameter.TIMEOUT.value: 10000,
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.1.254",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": InstrumentName.AGILENT_E5071B.value,
                "slot_id": 0,
            }
        ],
    }

    agilent_e5071b: dict[str, Any] = {
        "name": InstrumentName.AGILENT_E5071B,
        "alias": InstrumentName.AGILENT_E5071B.value,
        Parameter.POWER.value: -60.0,
    }

    instruments = [keysight_e5080b, agilent_e5071b]
    instrument_controllers = [
        keysight_e5080b_controller,
        agilent_e5071b_controller,
    ]

    buses: list[dict[str, Any]] = [
        {
            RUNCARD.ALIAS: "keysight_e5080b_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.AGILENT_E5071B.value],
            RUNCARD.CHANNELS: [0],
        },
        {
            RUNCARD.ALIAS: "agilent_e5071b_bus",
            RUNCARD.INSTRUMENTS: [InstrumentName.AGILENT_E5071B.value],
            RUNCARD.CHANNELS: [0],
        },
    ]

    runcard = {
        RUNCARD.NAME: name,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
        RUNCARD.BUSES: buses,
    }


class SauronQuantumMachines:
    """Test data of the sauron with quantum machines platform."""

    name = "sauron_quantum_machines"

    qmm = {
        "name": InstrumentName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}, "shareable": True},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [],
        "elements": [
            {
                "identifier": "drive_q0",
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 1},
                    "Q": {"controller": "con1", "port": 2},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0",
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 3},
                    "Q": {"controller": "con1", "port": 4},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "outputs": {"out1": {"controller": "con1", "port": 1}, "out2": {"controller": "con1", "port": 2}},
                "time_of_flight": 40,
                "smearing": 10,
                "threshold_rotation": 0.5,
                "threshold": 0.09,
                "intermediate_frequency": 6e9,
            },
            {"identifier": "flux_q0", "single_input": {"controller": "con1", "port": 5}},
        ],
        "run_octave_calibration": False,
    }

    qmm_with_octave = {
        "name": InstrumentName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_octave",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "connectivity": {"controller": "con1"},
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
                "rf_outputs": [
                    {"port": 1, "lo_frequency": 6e9},
                    {"port": 2, "lo_frequency": 6e9},
                    {"port": 3, "lo_frequency": 6e9},
                    {"port": 4, "lo_frequency": 6e9},
                    {"port": 5, "lo_frequency": 6e9},
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
        ],
        "run_octave_calibration": True,
    }

    qmm_with_octave_custom_connectivity = {
        "name": InstrumentName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_octave_custom_connectivity",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "rf_outputs": [
                    {
                        "port": 1,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 1},
                        "q_connection": {"controller": "con1", "port": 2},
                    },
                    {
                        "port": 2,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 3},
                        "q_connection": {"controller": "con1", "port": 4},
                    },
                    {
                        "port": 3,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 5},
                        "q_connection": {"controller": "con1", "port": 6},
                    },
                    {
                        "port": 4,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 7},
                        "q_connection": {"controller": "con1", "port": 8},
                    },
                    {
                        "port": 5,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 9},
                        "q_connection": {"controller": "con1", "port": 10},
                    },
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
                "if_outputs": [{"controller": "con1", "port": 1}, {"controller": "con1", "port": 2}],
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
        ],
        "run_octave_calibration": True,
    }

    qmm_with_opx1000 = {
        "name": InstrumentName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_opx1000",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "type": "opx1000",
                "fems": [
                    {
                        "fem": 1,
                        "shareable": True,
                        "analog_outputs": [
                            {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                            {"port": 2},
                            {"port": 3},
                            {"port": 4},
                            {"port": 5},
                            {"port": 6},
                            {"port": 7},
                            {"port": 8},
                        ],
                        "analog_inputs": [{"port": 1}, {"port": 2}],
                        "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
                    }
                ],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "connectivity": {"controller": "con1", "fem": 1},
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
                "rf_outputs": [
                    {
                        "port": 1,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 1},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 2},
                    },
                    {
                        "port": 2,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 3},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 4},
                    },
                    {
                        "port": 3,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 5},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 6},
                    },
                    {
                        "port": 4,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 7},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 8},
                    },
                    {
                        "port": 5,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 9},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 10},
                    },
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "fem": 1, "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
            {"identifier": "flux_q0", "single_input": {"controller": "con1", "fem": 1, "port": 5}},
        ],
        "run_octave_calibration": True,
    }

    qmm_controller = {
        "name": InstrumentControllerName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.111",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qmm",
                "slot_id": 0,
            }
        ],
    }

    qmm_with_octave_controller = {
        "name": InstrumentControllerName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_octave_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.111",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qmm_with_octave",
                "slot_id": 0,
            }
        ],
    }

    qmm_with_octave_custom_connectivity_controller = {
        "name": InstrumentControllerName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_octave_custom_connectivity_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.111",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qmm_with_octave_custom_connectivity",
                "slot_id": 0,
            }
        ],
    }

    qmm_with_opx1000_controller = {
        "name": InstrumentControllerName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_with_opx1000_controller",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.111",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "qmm_with_opx1000",
                "slot_id": 0,
            }
        ],
    }

    rohde_schwarz: dict[str, Any] = {
        "name": InstrumentName.ROHDE_SCHWARZ.value,
        "alias": "rohde_schwarz",
        Parameter.POWER.value: 15,
        Parameter.LO_FREQUENCY.value: 7.24730e09,
        Parameter.RF_ON.value: True,
    }

    qmm_controller_wrong_module = {
        "name": InstrumentControllerName.QUANTUM_MACHINES_CLUSTER.value,
        "alias": "qmm_controller_wrong_module",
        INSTRUMENTCONTROLLER.CONNECTION: {
            "name": ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.111",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rohde_schwarz",
                "slot_id": 0,
            }
        ],
    }

    instruments = [qmm, qmm_with_octave, qmm_with_octave_custom_connectivity, qmm_with_opx1000, rohde_schwarz]
    instrument_controllers = [
        qmm_controller,
        qmm_with_octave_controller,
        qmm_with_octave_custom_connectivity_controller,
        qmm_with_opx1000_controller,
        qmm_controller_wrong_module,
    ]

    buses: list[dict[str, Any]] = [
        {RUNCARD.ALIAS: "drive_q0", RUNCARD.INSTRUMENTS: ["qmm"], RUNCARD.CHANNELS: ["drive_q0"]},
        {RUNCARD.ALIAS: "readout_q0", RUNCARD.INSTRUMENTS: ["qmm"], RUNCARD.CHANNELS: ["readout_q0"]},
        {RUNCARD.ALIAS: "flux_q0", RUNCARD.INSTRUMENTS: ["qmm"], RUNCARD.CHANNELS: ["flux_q0"]},
        {RUNCARD.ALIAS: "drive_q0_rf", RUNCARD.INSTRUMENTS: ["qmm_with_octave"], RUNCARD.CHANNELS: ["drive_q0_rf"]},
        {RUNCARD.ALIAS: "readout_q0_rf", RUNCARD.INSTRUMENTS: ["qmm_with_octave"], RUNCARD.CHANNELS: ["readout_q0_rf"]},
        {
            RUNCARD.ALIAS: "drive_q0_rf_custom",
            RUNCARD.INSTRUMENTS: ["qmm_with_octave_custom_connectivity"],
            RUNCARD.CHANNELS: ["drive_q0_rf"],
        },
        {
            RUNCARD.ALIAS: "readout_q0_rf_custom",
            RUNCARD.INSTRUMENTS: ["qmm_with_octave_custom_connectivity"],
            RUNCARD.CHANNELS: ["readout_q0_rf"],
        },
        {
            RUNCARD.ALIAS: "drive_q0_opx1000",
            RUNCARD.INSTRUMENTS: ["qmm_with_opx1000"],
            RUNCARD.CHANNELS: ["drive_q0_rf"],
        },
        {
            RUNCARD.ALIAS: "readout_q0_opx1000",
            RUNCARD.INSTRUMENTS: ["qmm_with_opx1000"],
            RUNCARD.CHANNELS: ["readout_q0_rf"],
        },
    ]

    runcard = {
        RUNCARD.NAME: name,
        RUNCARD.BUSES: buses,
        RUNCARD.INSTRUMENTS: instruments,
        RUNCARD.INSTRUMENT_CONTROLLERS: instrument_controllers,
    }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vector Network Analyzer General Instrument Controller"""

from dataclasses import dataclass

from qililab.constants import DEFAULT_TIMEOUT
from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.typings.enums import ConnectionName
from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver


class VectorNetworkAnalyzerController(SingleInstrumentController):
    """Vector Network Analyzer General Instrument Controller

    Args:
        settings (VectorNetworkAnalyzerControllerSettings): Settings of the instrument controller.
    """

    @dataclass
    class VectorNetworkAnalyzerControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific VectorNetworkAnalyzer Controller."""

        timeout: float = DEFAULT_TIMEOUT

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: VectorNetworkAnalyzerControllerSettings
    device: VectorNetworkAnalyzerDriver

    @property
    def timeout(self):
        """VectorNetworkAnalyzer 'timeout' property.

        Returns:
            float: settings.timeout.
        """
        return self.settings.timeout

    @timeout.setter
    def timeout(self, value: float):
        """sets the timeout"""
        self.settings.timeout = value
        self.device.set_timeout(value=self.settings.timeout)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Attenuator class."""

from dataclasses import dataclass

from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings.instruments.mini_circuits import MiniCircuitsDriver


@InstrumentFactory.register
class Attenuator(Instrument):
    """Attenuator class.

    Args:
        name (InstrumentName): name of the instrument
        device (MiniCircuitsDriver): Instance of the MiniCircuitsDriver class.
        settings (StepAttenuatorSettings): Settings of the instrument.
    """

    name = InstrumentName.MINI_CIRCUITS

    @dataclass
    class StepAttenuatorSettings(Instrument.InstrumentSettings):
        """Step attenuator settings."""

        attenuation: float

    settings: StepAttenuatorSettings
    device: MiniCircuitsDriver

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set instrument settings."""
        if parameter == Parameter.ATTENUATION:
            self.settings.attenuation = float(value)
            if self.is_device_active():
                self.device.setup(attenuation=self.attenuation)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Set instrument settings."""
        if parameter == Parameter.ATTENUATION:
            return self.attenuation
        raise ParameterNotFound(self, parameter)

    @check_device_initialized
    def initial_setup(self):
        """performs an initial setup."""
        self.device.setup(attenuation=self.attenuation)

    @check_device_initialized
    def turn_off(self):
        """Turn off an instrument."""

    @check_device_initialized
    def turn_on(self):
        """Turn on an instrument."""

    @check_device_initialized
    def reset(self):
        """Reset instrument."""

    @property
    def attenuation(self):
        """Attenuator 'attenuation' property.

        Returns:
            float: Attenuation.
        """
        return self.settings.attenuation



"""Unit tests for the `about` function."""
import io
import platform
import sys
from subprocess import check_output

import pyvisa_py
import qblox_instruments
import qcodes
import qcodes_contrib_drivers
import qibo
import qpysequence
from qm.version import __version__ as qm_version

import qililab as ql


def test_about():
    """Test that the `about` function prints the correct information."""
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput  # Redirect output
    ql.about()
    sys.stdout = sys.__stdout__  # Reset redirect

    expected_string = f"""Platform info:             {platform.platform(aliased=True)}
Python version:            {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}
PyVISA version:            {pyvisa_py.__version__}
QCodes version:            {qcodes.__version__}
QCodes Contrib version:    {qcodes_contrib_drivers.__version__}
Qblox Instrument version:  {qblox_instruments.__version__}
Qpysequence version:       {qpysequence.__version__}
Quantum Machines version:  {qm_version}
Qibo version:              {qibo.__version__}
"""

    assert expected_string == capturedOutput.getvalue()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to interface with the voltage source Qblox S4g
"""

from dataclasses import dataclass
from time import sleep
from typing import Any

from qililab.config import logger
from qililab.instruments.current_source import CurrentSource
from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings import QbloxS4g as QbloxS4gDriver


@InstrumentFactory.register
class QbloxS4g(CurrentSource):
    """Qblox S4g class

    Args:
        name (InstrumentName): name of the instrument
        device (Qblox_S4g): Instance of the qcodes S4g class.
        settings (QbloxS4gSettings): Settings of the instrument.
    """

    name = InstrumentName.QBLOX_S4G

    @dataclass
    class QbloxS4gSettings(CurrentSource.CurrentSourceSettings):
        """Contains the settings of a specific signal generator."""

    settings: QbloxS4gSettings
    device: QbloxS4gDriver

    def dac(self, dac_index: int):
        """get channel associated to the specific dac

        Args:
            dac_index (int): channel index

        Returns:
            _type_: _description_
        """
        return getattr(self.device, f"dac{dac_index}")

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set Qblox instrument calibration settings."""

        if channel_id is None:
            if len(self.dacs) == 1:
                channel_id = self.dacs[0]
            else:
                raise ValueError(f"channel not specified to update instrument {self.name.value}")

        channel_id = int(channel_id)
        if channel_id > 3:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 4 -> maximum channel_id should be 3."
            )

        channel = self.dac(dac_index=channel_id) if self.is_device_active() else None

        if parameter == Parameter.CURRENT:
            self._set_current(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.SPAN:
            self._set_span(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_ENABLED:
            self._set_ramping_enabled(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_RATE:
            self._set_ramping_rate(value=value, channel_id=channel_id, channel=channel)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if channel_id is None:
            if len(self.dacs) == 1:
                channel_id = self.dacs[0]
            else:
                raise ValueError(f"channel not specified to update instrument {self.name.value}")

        channel_id = int(channel_id)
        if channel_id > 3:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 4 -> maximum channel_id should be 3."
            )
        if hasattr(self.settings, parameter.value):
            index = self.dacs.index(channel_id)
            return getattr(self.settings, parameter.value)[index]
        raise ParameterNotFound(self, parameter)

    def _set_current(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the current"""
        index = self.dacs.index(channel_id)
        self.settings.current[index] = float(value)

        if self.is_device_active():
            channel.current(self.current[index])

    def _set_span(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the span"""
        index = self.dacs.index(channel_id)
        self.settings.span[index] = str(value)

        if self.is_device_active():
            channel.span(self.span[index])

    def _set_ramping_enabled(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramping_enabled"""
        index = self.dacs.index(channel_id)
        self.settings.ramping_enabled[index] = bool(value)

        if self.is_device_active():
            channel.ramping_enabled(self.ramping_enabled[index])

    def _set_ramping_rate(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramp_rate"""
        index = self.dacs.index(channel_id)
        self.settings.ramp_rate[index] = float(value)

        if self.is_device_active():
            channel.ramp_rate(self.ramp_rate[index])

    @check_device_initialized
    def initial_setup(self):
        """performs an initial setup."""
        for channel_id in self.dacs:
            index = self.dacs.index(channel_id)
            channel = self.dac(dac_index=channel_id)
            channel.ramping_enabled(self.ramping_enabled[index])
            channel.ramp_rate(self.ramp_rate[index])
            channel.span(self.span[index])
            channel.current(self.current[index])
            logger.debug("SPI current set to %f", channel.current())
            while channel.is_ramping():
                sleep(0.1)

    @check_device_initialized
    def turn_on(self):
        """Dummy method."""

    @check_device_initialized
    def turn_off(self):
        """Stop outputing current."""
        for channel_id in self.settings.dacs:
            channel = self.dac(dac_index=channel_id)
            channel.current(0)

    @check_device_initialized
    def reset(self):
        """Reset instrument."""
        for channel_id in self.settings.dacs:
            channel = self.dac(dac_index=channel_id)
            channel.current(0)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DataUnavailable Exception class"""


class DataUnavailable(Exception):
    """Exception raised when a method is requesting data that is not available.

    Args:
        message (str): Optional message to be displayed
    """

    def __init__(self, message: str | None):
        super().__init__("Requesting data not available" if message is None else message)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Execute function used to execute a qibo Circuit using the given runcard."""

from qibo.models import Circuit
from tqdm.auto import tqdm

from qililab.digital import DigitalTranspilationConfig
from qililab.result import Result

from .data_management import build_platform


def execute(
    program: Circuit | list[Circuit],
    runcard: str | dict,
    nshots: int = 1,
    transpilation_config: DigitalTranspilationConfig | None = None,
) -> Result | list[Result]:
    """Executes a Qibo circuit (or a list of circuits) with qililab and returns the results.

    The ``program`` argument is first translated into pulses using the transpilation settings of the runcard and the passed transpile
    configuration. Then the pulse will be compiled into the runcard machines assembly programs, and executed.

    The transpilation is performed using the :meth:`.CircuitTranspiler.transpile_circuit()` method. Refer to the method's documentation or :ref:`Transpilation <transpilation>` for more detailed information.

    The main stages of this process are: **1.** Routing, **2.** Canceling Hermitian pairs, **3.** Translate to native gates, **4.** Correcting Drag phases, **5** Optimize Drag gates, **6.** Convert to pulse schedule.

    .. note ::

        Default steps are only: **3.**, **4.**, and **6.**, since they are always needed.

        To do Step **1.** set routing=True in transpilation_config (default behavior skips it).

        To do Steps **2.** and **5.** set optimize=True in transpilation_config (default behavior skips it)

    Args:
        circuit (Circuit | list[Circuit]): Qibo Circuit.
        runcard (str | dict): If a string, path to the YAML file containing the serialization of the Platform to be
            used. If a dictionary, the serialized platform to be used.
        nshots (int, optional): Number of shots to execute. Defaults to 1.
        transpilation_config (DigitalTranspilationConfig, optional): :class:`.DigitalTranspilationConfig` dataclass containing
            the configuration used during transpilation. Defaults to ``None`` (not changing any default value).
            Check the class:`.DigitalTranspilationConfig` documentation for the keys and values it can contain.

    Returns:
        Result | list[Result]: :class:`Result` class (or list of :class:`Result` classes) containing the results of the
            execution.

    |

    Example Usage:

    .. code-block:: python

        import numpy as np
        from qibo import Circuit, gates
        from qibo.transpiler import Sabre, ReverseTraversal
        from qililab.digital import DigitalTranspilationConfig
        import qililab as ql

        # Create circuit:
        nqubits = 5
        c = Circuit(nqubits)
        for qubit in range(nqubits):
            c.add(gates.H(qubit))
        c.add(gates.CNOT(2, 0))
        c.add(gates.RY(4, np.pi / 4))
        c.add(gates.X(3))
        c.add(gates.M(*range(3)))
        c.add(gates.SWAP(4, 2))
        c.add(gates.RX(1, 3 * np.pi / 2))

        # Create transpilation config:
        transpilation = DigitalTranspilationConfig(routing=True, optimize=False, router=Sabre, placer=ReverseTraversal)

        # Execute with automatic transpilation:
        probabilities = ql.execute(c, runcard="./runcards/galadriel.yml", transpilation_config=transpilation)
    """
    if isinstance(program, Circuit):
        program = [program]

    # Initialize platform and connect to the instruments
    platform = build_platform(runcard=runcard)
    platform.connect()

    try:
        platform.initial_setup()
        platform.turn_on_instruments()
        results = [
            # Execute circuit
            platform.execute(
                circuit,
                num_avg=1,
                repetition_duration=200_000,
                num_bins=nshots,
                transpilation_config=transpilation_config,
            )
            for circuit in tqdm(program, total=len(program))
        ]
        platform.disconnect()
        return results[0] if len(results) == 1 else results
    except Exception as e:
        platform.disconnect()
        raise e



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qililab.qprogram.operations.operation import Operation
from qililab.waveforms import IQPair, Waveform
from qililab.yaml import yaml


@yaml.register_class
class Measure(Operation):
    def __init__(
        self,
        bus: str,
        waveform: IQPair,
        weights: IQPair,
        save_adc: bool = False,
        rotation: float | None = None,
        demodulation: bool = True,
    ) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: IQPair = waveform
        self.weights: IQPair = weights
        self.save_adc: bool = save_adc
        self.rotation: float | None = rotation
        self.demodulation: bool = demodulation

    def get_waveforms(self) -> tuple[Waveform, Waveform]:
        """Get the waveforms.

        Returns:
            tuple[Waveform, Waveform | None]: The waveforms as tuple.
        """
        wf_I: Waveform = self.waveform.I
        wf_Q: Waveform = self.waveform.Q
        return wf_I, wf_Q


@yaml.register_class
class MeasureWithCalibratedWaveform(Operation):
    def __init__(
        self,
        bus: str,
        waveform: str,
        weights: IQPair,
        save_adc: bool = False,
        rotation: float | None = None,
        demodulation: bool = True,
    ) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: str = waveform
        self.weights: IQPair = weights
        self.save_adc: bool = save_adc
        self.rotation: float | None = rotation
        self.demodulation: bool = demodulation


@yaml.register_class
class MeasureWithCalibratedWeights(Operation):
    def __init__(
        self,
        bus: str,
        waveform: IQPair,
        weights: str,
        save_adc: bool = False,
        rotation: float | None = None,
        demodulation: bool = True,
    ) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: IQPair = waveform
        self.weights: str = weights
        self.save_adc: bool = save_adc
        self.rotation: float | None = rotation
        self.demodulation: bool = demodulation


@yaml.register_class
class MeasureWithCalibratedWaveformWeights(Operation):
    def __init__(
        self,
        bus: str,
        waveform: str,
        weights: str,
        save_adc: bool = False,
        rotation: float | None = None,
        demodulation: bool = True,
    ) -> None:
        super().__init__()
        self.bus: str = bus
        self.waveform: str = waveform
        self.weights: str = weights
        self.save_adc: bool = save_adc
        self.rotation: float | None = rotation
        self.demodulation: bool = demodulation



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable

from qililab.analog.fluqe_parameter import FluqeParameter


class Qubit2LevelTranspiler:
    """Implementation of the transpiler for the 2 level qubit. This is done mainly by inverting the same functions
    used in the single_qubit_2level emulator model.

    Args:
        epsilon_model (Callable): epsilon model
        delta_model (Callable): delta model

    Calling an instance of this class returns the fluxes phix, phiz for some given Delta, epsilon.
    """

    def __init__(self, epsilon_model: Callable, delta_model: Callable):
        """Init method. See class description for more details"""

        self.delta_model = delta_model
        self.epsilon_model = epsilon_model

        # Magnetic energy bias (epsilon)
        self.epsilon = FluqeParameter(name="epsilon", set_method=self._set_epsilon)
        # Qubit gap (delta)
        self.delta = FluqeParameter(name="delta", set_method=self._set_delta)
        # flux parameters
        self.phiz = FluqeParameter(name="phiz")
        self.phix = FluqeParameter(name="phix")

    def __call__(self, delta: float, epsilon: float) -> tuple[Any, Any]:
        """Transpiles Delta and Epsilon to phix, phiz"""
        self.delta(delta)
        self.epsilon(epsilon)
        return self.phix(), self.phiz()

    def _set_delta(self, delta):
        # sets the value of delta via raw and updates phix accordingly
        self.phix.set_raw(self.delta_model(delta))
        return delta

    def _set_epsilon(self, epsilon):
        """sets the value of epsilon via raw and updates phiz accordingly. Does not update phix
        since this is meant to be used in the transpiler in conjunction with setting delta (which
        already updates phix)
        """
        self.phiz.set_raw(self.epsilon_model(phix=self.phix(), epsilon=epsilon))
        return epsilon



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BusElement class"""

from dataclasses import asdict

from qililab.constants import RUNCARD
from qililab.settings.settings import Settings
from qililab.typings.factory_element import FactoryElement
from qililab.utils import dict_factory


class BusElement(FactoryElement):
    """Class BusElement. All bus element classes must inherit from this class."""

    settings: Settings

    def to_dict(self):
        """Return a dict representation of the BusElement class."""
        return {RUNCARD.NAME: self.name.value} | asdict(self.settings, dict_factory=dict_factory)

    def short_dict(self):
        """Return a dict representation of the BusElement class discarding all static elements."""
        return {key: value for key, value in self.to_dict().items() if key not in [RUNCARD.NAME]}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""

from .decorators import check_device_initialized, log_set_parameter
from .instrument import Instrument, ParameterNotFound
from .instruments import Instruments
from .mini_circuits import Attenuator
from .quantum_machines import QuantumMachinesCluster
from .rohde_schwarz import SGS100A
from .signal_generator import SignalGenerator
from .utils import InstrumentFactory

__all__ = [
    "SGS100A",
    "Attenuator",
    "Instrument",
    "InstrumentFactory",
    "Instruments",
    "ParameterNotFound",
    "QuantumMachinesCluster",
    "SignalGenerator",
    "check_device_initialized",
    "log_set_parameter",
]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bin Data"""

from dataclasses import dataclass

import numpy as np

from qililab.utils.nested_data_class import nested_dataclass


@nested_dataclass
class BinsData:
    """Holds the integrated and thresholded values of all bins in a sequencer."""

    @dataclass
    class QbloxIntegrationData:
        """QbloxIntegrationData class."""

        path0: list[float]
        path1: list[float]

        def __post_init__(self):
            """Remove nan values."""  # FIXME: Since we cannot do ascending loops in Qpysequence, we need to
            # use always a number of bins = num_loops + 1. Thus the first bin is always a nan.
            self.path0 = [value for value in self.path0 if not np.isnan(value)]
            self.path1 = [value for value in self.path1 if not np.isnan(value)]

    integration: QbloxIntegrationData
    threshold: list
    avg_cnt: list

    def __post_init__(self):
        """Remove nan values."""  # FIXME: Since we cannot do ascending loops in Qpysequence, we need to
        # use always a number of bins = num_loops + 1. Thus the first bin is always a nan.
        self.threshold = [value for value in self.threshold if not np.isnan(value)]
        self.avg_cnt = [value for value in self.avg_cnt if not np.isnan(value)]

    def __len__(self) -> int:
        """Returns the length of the QbloxIntegrationData, corresponding to the number of bins.

        Returns:
            int: Length of the QbloxIntegrationData.
        """
        return len(self.threshold)



"""This file tests the the ``qmm_controller`` class"""
import pytest

from qililab.instrument_controllers.quantum_machines import QuantumMachinesClusterController
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.platform import Platform
from qililab.settings import Settings
from tests.data import SauronQuantumMachines
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronQuantumMachines.runcard)


class TestQuantumMachinesClusterController:
    """This class contains the unit tests for the ``QuantumMachinesClusterController`` class."""

    @pytest.mark.parametrize("controller_alias", ["qmm_controller", "qmm_with_octave_controller"])
    def test_initialization(self, platform: Platform, controller_alias: str):
        """Test QuantumMachinesClusterController has been initialized correctly"""
        controller_instance = platform.get_element(alias=controller_alias)
        assert isinstance(controller_instance, QuantumMachinesClusterController)

        controller_settings = controller_instance.settings
        assert isinstance(controller_settings, Settings)

        controller_modules = controller_instance.modules
        assert len(controller_modules) == 1
        assert isinstance(controller_modules[0], QuantumMachinesCluster)



import numpy as np
import pytest
import math

from qililab.waveforms import Arbitrary

samples = np.concatenate([np.linspace(0, 10_000, 1000), np.linspace(10_000, 0, 1000)])


@pytest.fixture(name="arbitrary")
def fixture_square():
    return Arbitrary(samples=samples)


class TestArbitrary:
    def test_init(self, arbitrary):
        # test init method
        assert np.allclose(arbitrary.samples, samples)

    def test_envelope(self, arbitrary):
        # test envelope method
        assert np.allclose(arbitrary.envelope(), samples)

    def test_envelope_with_higher_resolution(self, arbitrary):
        # test envelope method
        resolution = 100
        envelope = arbitrary.envelope(resolution=resolution)
        assert math.isclose(envelope[0], samples[0], abs_tol=1e-9)
        assert math.isclose(envelope[-1], samples[-1], abs_tol=1e-9)
        assert max(envelope) == max(samples)
        assert min(envelope) == min(envelope)
        assert np.allclose(envelope, np.array([
            0.00000000e+00, 1.11111111e+03, 2.22222222e+03, 3.33333333e+03,
            4.44444444e+03, 5.55555556e+03, 6.66666667e+03, 7.77777778e+03,
            8.88888889e+03, 1.00000000e+04, 1.00000000e+04, 8.88888889e+03,
            7.77777778e+03, 6.66666667e+03, 5.55555556e+03, 4.44444444e+03,
            3.33333333e+03, 2.22222222e+03, 1.11111111e+03, 0.00000000e+00
        ]))

    def test_envelope_with_constant_values(self):
        resolution = 20_000
        arbitrary = Arbitrary(samples=np.ones(resolution))
        envelope = arbitrary.envelope(resolution=resolution)
        assert np.allclose(envelope, np.ones(200))

    def test_envelope_with_higher_resolution_raises_error(self, arbitrary):
        resolution = 20_000
        with pytest.raises(ValueError):
            _ = arbitrary.envelope(resolution=resolution)

    def test_get_duration_method(self, arbitrary):
        assert arbitrary.get_duration() == 2000



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the methods and classes used to define a pulse sequence.

.. currentmodule:: qililab.pulse

Pulse Schedules
~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~PulseSchedule
    ~PulseBusSchedule
    ~PulseEvent

Pulses
~~~~~~

.. autosummary::
    :toctree: api

    ~Pulse

Pulse Shapes
=============

.. autosummary::
    :toctree: api

    ~PulseShape
    ~Gaussian
    ~Drag
    ~Rectangular
    ~SNZ
    ~Cosine

Distortions
~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~PulseDistortion
    ~LFilterCorrection
    ~BiasTeeCorrection
    ~ExponentialCorrection
"""

from .pulse import Pulse
from .pulse_bus_schedule import PulseBusSchedule
from .pulse_distortion import BiasTeeCorrection, ExponentialCorrection, LFilterCorrection, PulseDistortion
from .pulse_event import PulseEvent
from .pulse_schedule import PulseSchedule
from .pulse_shape import SNZ, Cosine, Drag, FlatTop, Gaussian, PulseShape, Rectangular
from .qblox_compiler import QbloxCompiler

__all__ = [
    "SNZ",
    "BiasTeeCorrection",
    "Cosine",
    "Drag",
    "ExponentialCorrection",
    "FlatTop",
    "Gaussian",
    "LFilterCorrection",
    "Pulse",
    "PulseBusSchedule",
    "PulseDistortion",
    "PulseEvent",
    "PulseSchedule",
    "PulseShape",
    "QbloxCompiler",
    "Rectangular",
]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox module class"""

from dataclasses import dataclass
from typing import ClassVar, Sequence

from qpysequence import Sequence as QpySequence

from qililab.config import logger
from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.qblox.qblox_sequencer import QbloxSequencer
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.typings import ChannelID, Parameter, ParameterValue
from qililab.typings.instruments import QcmQrm


class QbloxModule(Instrument):
    """Qblox Module class.

    Args:
        device (QcmQrm): Instance of the Qblox QcmQrm class used to connect to the instrument.
        settings (QbloxPulsarSettings): Settings of the instrument.
    """

    _MAX_BINS: int = 131072
    _NUM_MAX_SEQUENCERS: int = 6
    _NUM_MAX_AWG_OUT_CHANNELS: int = 4
    _MIN_WAIT_TIME: int = 4  # in ns

    @dataclass
    class QbloxModuleSettings(Instrument.InstrumentSettings):
        """Contains the settings of a specific module.

        Args:
            awg_sequencers (Sequence[QbloxSequencer]): list of settings for each sequencer
            out_offsets (list[float]): list of offsets for each output of the qblox module
        """

        awg_sequencers: Sequence[QbloxSequencer]
        out_offsets: list[float]

        def __post_init__(self):
            """build QbloxSequencer"""
            if len(self.awg_sequencers) > QbloxModule._NUM_MAX_SEQUENCERS:
                raise ValueError(
                    f"The number of sequencers must be less or equal than {QbloxModule._NUM_MAX_SEQUENCERS}. Received: {len(self.awg_sequencers)}"
                )

            self.awg_sequencers = [
                (QbloxSequencer(**sequencer) if isinstance(sequencer, dict) else sequencer)
                for sequencer in self.awg_sequencers
            ]
            super().__post_init__()

    settings: QbloxModuleSettings
    device: QcmQrm
    # Cache containing the last compiled pulse schedule for each sequencer
    cache: ClassVar[dict[int, PulseBusSchedule]] = {}

    def __init__(self, settings: dict):
        # The sequences dictionary contains all the compiled sequences for each sequencer. Sequences are saved and handled at the compiler
        self.sequences: dict[int, QpySequence] = {}  # {sequencer_idx: (program), ...}
        self.num_bins: int = 1
        super().__init__(settings=settings)

    @property
    def num_sequencers(self):
        """Number of sequencers in the AWG

        Returns:
            int: number of sequencers
        """
        return len(self.settings.awg_sequencers)

    @property
    def awg_sequencers(self):
        """AWG 'awg_sequencers' property."""
        return self.settings.awg_sequencers

    def get_sequencer(self, sequencer_id: int) -> QbloxSequencer:
        """Get sequencer from the sequencer identifier

        Args:
            sequencer_id (int): sequencer identifier

        Returns:
            AWGSequencer: sequencer associated with the sequencer_id
        """
        for sequencer in self.awg_sequencers:
            if sequencer.identifier == sequencer_id:
                return sequencer
        raise IndexError(f"There is no sequencer with id={sequencer_id}.")

    @check_device_initialized
    def initial_setup(self):
        """Initial setup"""
        self._map_connections()
        self.clear_cache()
        for sequencer in self.awg_sequencers:
            sequencer_id = sequencer.identifier
            # Set `sync_en` flag to False (this value will be set to True if the sequencer is used in the execution)
            self.device.sequencers[sequencer_id].sync_en(False)
            self.device.sequencers[sequencer_id].marker_ovr_en(False)
            self._set_nco(sequencer_id=sequencer_id)
            self._set_gain_i(value=sequencer.gain_i, sequencer_id=sequencer_id)
            self._set_gain_q(value=sequencer.gain_q, sequencer_id=sequencer_id)
            self._set_offset_i(value=sequencer.offset_i, sequencer_id=sequencer_id)
            self._set_offset_q(value=sequencer.offset_q, sequencer_id=sequencer_id)
            self._set_hardware_modulation(value=sequencer.hardware_modulation, sequencer_id=sequencer_id)
            self._set_gain_imbalance(value=sequencer.gain_imbalance, sequencer_id=sequencer_id)
            self._set_phase_imbalance(value=sequencer.phase_imbalance, sequencer_id=sequencer_id)

        for idx, offset in enumerate(self.out_offsets):
            self._set_out_offset(output=idx, value=offset)

    def sync_sequencer(self, sequencer_id: int) -> None:
        """Syncs all sequencers."""
        sequencer = self.get_sequencer(sequencer_id=sequencer_id)
        self.device.sequencers[sequencer.identifier].sync_en(True)

    def desync_sequencer(self, sequencer_id: int) -> None:
        """Syncs all sequencers."""
        sequencer = self.get_sequencer(sequencer_id=sequencer_id)
        self.device.sequencers[sequencer.identifier].sync_en(False)

    def desync_sequencers(self) -> None:
        """Desyncs all sequencers."""
        for sequencer in self.awg_sequencers:
            self.device.sequencers[sequencer.identifier].sync_en(False)

    def set_markers_override_enabled(self, value: bool, sequencer_id: int):
        """Set markers override flag ON/OFF for the sequencers associated with port."""
        sequencer = self.get_sequencer(sequencer_id=sequencer_id)
        self.device.sequencers[sequencer.identifier].marker_ovr_en(value)

    def set_markers_override_value(self, value: int, sequencer_id: int):
        """Set markers override value for all sequencers."""
        sequencer = self.get_sequencer(sequencer_id=sequencer_id)
        self.device.sequencers[sequencer.identifier].marker_ovr_value(value)

    def module_type(self):
        """returns the qblox module type. Options: QCM or QRM"""
        return self.device.module_type()

    def run(self, channel_id: ChannelID):
        """Run the uploaded program"""
        sequencer = next((sequencer for sequencer in self.awg_sequencers if sequencer.identifier == channel_id), None)
        if sequencer is not None and sequencer.identifier in self.sequences:
            self.device.arm_sequencer(sequencer=sequencer.identifier)
            self.device.start_sequencer(sequencer=sequencer.identifier)

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None) -> None:
        """Set Qblox instrument calibration settings."""
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            self._set_out_offset(output=output, value=value)
            return

        if channel_id is None:
            raise Exception(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        channel_id = int(channel_id)
        if parameter == Parameter.GAIN:
            self._set_gain(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_I:
            self._set_gain_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_Q:
            self._set_gain_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_I:
            self._set_offset_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_Q:
            self._set_offset_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.IF:
            self._set_frequency(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.HARDWARE_MODULATION:
            self._set_hardware_modulation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_IMBALANCE:
            self._set_gain_imbalance(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.PHASE_IMBALANCE:
            self._set_phase_imbalance(value=value, sequencer_id=channel_id)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            return self.out_offsets[output]

        if channel_id is None:
            raise Exception(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        channel_id = int(channel_id)
        sequencer = self.get_sequencer(sequencer_id=channel_id)

        if parameter == Parameter.GAIN:
            return sequencer.gain_i, sequencer.gain_q

        if hasattr(sequencer, parameter.value):
            return getattr(sequencer, parameter.value)

        raise ParameterNotFound(self, parameter)

    def _set_hardware_modulation(self, value: float | str | bool, sequencer_id: int):
        """set hardware modulation

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        self.get_sequencer(sequencer_id=sequencer_id).hardware_modulation = bool(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].mod_en_awg(bool(value))

    def _set_frequency(self, value: float | str | bool, sequencer_id: int):
        """set frequency

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self.get_sequencer(sequencer_id=sequencer_id).intermediate_frequency = float(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].nco_freq(float(value))

    def _set_offset_i(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self.get_sequencer(sequencer_id=sequencer_id).offset_i = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "offset_awg_path0")(float(value))

    def _set_offset_q(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self.get_sequencer(sequencer_id=sequencer_id).offset_q = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "offset_awg_path1")(float(value))

    def _set_out_offset(self, output: int, value: float | str | bool):
        """Set output offsets of the Qblox device.

        Args:
            output (int): output to update
            value (float | str | bool): value to update

        Raises:
            ValueError: when value type is not float or int
        """
        if output > len(self.out_offsets):
            raise IndexError(
                f"Output {output} is out of range. The runcard has only {len(self.out_offsets)} output offsets defined."
                " Please update the list of output offsets of the runcard such that it contains a value for each "
                "output of the device."
            )
        self.out_offsets[output] = value

        if self.is_device_active():
            getattr(self.device, f"out{output}_offset")(float(value))

    def _set_gain_i(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self.get_sequencer(sequencer_id=sequencer_id).gain_i = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "gain_awg_path0")(float(value))

    def _set_gain_q(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self.get_sequencer(sequencer_id=sequencer_id).gain_q = float(value)
        # update value in the instrument
        if self.is_device_active():
            sequencer = self.device.sequencers[sequencer_id]
            getattr(sequencer, "gain_awg_path1")(float(value))

    def _set_gain(self, value: float | str | bool, sequencer_id: int):
        """set gain

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._set_gain_i(value=value, sequencer_id=sequencer_id)
        self._set_gain_q(value=value, sequencer_id=sequencer_id)

    @check_device_initialized
    def turn_off(self):
        """Stop the QBlox sequencer from sending pulses."""
        for seq_idx in range(self.num_sequencers):
            self.device.stop_sequencer(sequencer=seq_idx)

    @check_device_initialized
    def turn_on(self):
        """Turn on an instrument."""

    def clear_cache(self):
        """Empty cache."""
        self.cache = {}
        self.sequences = {}

    @check_device_initialized
    def reset(self):
        """Reset instrument."""
        self.clear_cache()
        self.device.reset()

    def upload_qpysequence(self, qpysequence: QpySequence, channel_id: ChannelID):
        """Upload the qpysequence to its corresponding sequencer.

        Args:
            qpysequence (QpySequence): The qpysequence to upload.
            port (str): The port of the sequencer to upload to.
        """
        sequencer = next((sequencer for sequencer in self.awg_sequencers if sequencer.identifier == channel_id), None)
        if sequencer is not None:
            logger.info("Sequence program: \n %s", repr(qpysequence._program))
            self.device.sequencers[sequencer.identifier].sequence(qpysequence.todict())
            self.sequences[sequencer.identifier] = qpysequence

    def upload(self, channel_id: ChannelID):
        """Upload all the previously compiled programs to its corresponding sequencers.

        This method must be called after the method ``compile`` in the compiler
        Args:
            port (str): The port of the sequencer to upload to.
        """
        sequencer = next((sequencer for sequencer in self.awg_sequencers if sequencer.identifier == channel_id), None)
        if sequencer is not None and sequencer.identifier in self.sequences:
            sequence = self.sequences[sequencer.identifier]
            logger.info("Uploaded sequence program: \n %s", repr(sequence._program))  # pylint: disable=protected-access
            self.device.sequencers[sequencer.identifier].sequence(sequence.todict())
            self.device.sequencers[sequencer.identifier].sync_en(True)

    def _set_nco(self, sequencer_id: int):
        """Enable modulation of pulses and setup NCO frequency."""
        if self.get_sequencer(sequencer_id=sequencer_id).hardware_modulation:
            self._set_hardware_modulation(
                value=self.get_sequencer(sequencer_id=sequencer_id).hardware_modulation, sequencer_id=sequencer_id
            )
            self._set_frequency(
                value=self.get_sequencer(sequencer_id=sequencer_id).intermediate_frequency, sequencer_id=sequencer_id
            )

    def _set_gain_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q gain imbalance of sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """

        self.get_sequencer(sequencer_id=sequencer_id).gain_imbalance = float(value)

        if self.is_device_active():
            self.device.sequencers[sequencer_id].mixer_corr_gain_ratio(float(value))

    def _set_phase_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q phase imbalance of sequencer.

         Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self.get_sequencer(sequencer_id=sequencer_id).phase_imbalance = float(value)
        if self.is_device_active():
            self.device.sequencers[sequencer_id].mixer_corr_phase_offset_degree(float(value))

    def _map_connections(self):
        """Disable all connections and map sequencer paths with output channels."""
        # Disable all connections
        self.device.disconnect_outputs()

        for sequencer_dataclass in self.awg_sequencers:
            sequencer = self.device.sequencers[sequencer_dataclass.identifier]
            for path, output in zip(["I", "Q"], sequencer_dataclass.outputs):
                getattr(sequencer, f"connect_out{output}")(path)

    @property
    def out_offsets(self):
        """Returns the offsets of each output of the qblox module."""
        return self.settings.out_offsets



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the methods and classes used for automatic calibration.

A high level view of how the automatic calibration algorithm works and how the different classes and methods are used
can be read from: https://arxiv.org/abs/1803.03226

**Examples** are found in the `qililab/examples/calibration` folder

.. currentmodule:: qililab.calibration

CalibrationController Class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~CalibrationController

CalibrationNode Class
~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~CalibrationNode

Calibration-related methods
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~export_nb_outputs
"""

from .calibration_controller import CalibrationController
from .calibration_node import CalibrationNode, export_nb_outputs

__all__ = ["CalibrationController", "CalibrationNode", "export_nb_outputs"]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from qililab.typings.enums import Parameter


@dataclass
class GateEventSettings:
    """Settings for a single gate event. A gate event is an element of a gate schedule, which is the
    sequence of gate events that define what a gate does (the pulse events it consists of).

    A gate event is made up of GatePulseSettings, which contains pulse specific information, and extra arguments
    (bus and wait_time) which are related to the context in which the specific pulse in GatePulseSettings is applied

    Attributes:
        bus (str): bus through which the pulse is to be sent. The string has to match that of the bus alias in the runcard
        pulse (GatePulseSettings): settings of the bus to be launched
        wait_time (int): time to wait w.r.t gate start time (taken as 0) before launching the pulse
    """

    @dataclass
    class GatePulseSettings:
        """Single pulse settings

        Attributes:
            amplitude (float): amplitude of the pulse
            phase (float): phase of the pulse
            duration (int): pulse duration
            shape (dict): pulse envelope
            options (dict): optional gate dependant parameters
        """

        amplitude: float
        phase: float
        duration: int
        shape: dict
        options: dict | None = None

    bus: str
    pulse: GatePulseSettings
    wait_time: int = 0

    def __post_init__(self):
        """post init method to initialize pulse settings from runcard yaml file"""
        self.pulse = self.GatePulseSettings(**self.pulse)

    def set_parameter(self, parameter: Parameter, value: float | str | bool):
        """Change a given parameter from settings. Will look up into subclasses.

        Args:
            parameter (Parameter): parameter to be set
            value (float | str | bool): value of the parameter
        """
        param = parameter.value
        if hasattr(self, param):
            setattr(self, param, value)
        elif hasattr(self.pulse, param):
            setattr(self.pulse, param, value)
        else:
            self.pulse.shape[param] = value

    def get_parameter(self, parameter: Parameter):
        """Get a parameter from settings. Will look up into subclasses.

        Args:
            parameter (Parameter): Parameter to get.
        """
        param = parameter.value
        if hasattr(self, param):
            return getattr(self, param)
        if hasattr(self.pulse, param):
            return getattr(self.pulse, param)
        return self.pulse.shape[param]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MiniCircuits driver."""

import urllib
import urllib.error
from dataclasses import dataclass

from qililab.typings.instruments.device import Device


@dataclass
class MiniCircuitsDriver(Device):
    """Typing class of the driver for the MiniCircuits instrument.

    Args:
        name (str): Name of the instrument
        address (str): IP address of the instrument
    """

    name: str
    address: str

    def __post_init__(self):
        """Initialize the driver."""
        self._http_request(command="MN?")

    def setup(self, attenuation: float):
        """Set instrument settings."""
        self._http_request(command=f"SETATT={attenuation}")

    def _http_request(self, command: str):
        """Send an HTTP request with the given command.

        Args:
            command (str): Command to send via HTTP.
        """
        try:
            request = urllib.request.Request(f"http://{self.address}/:{command}")  # type: ignore
            with urllib.request.urlopen(request, timeout=2) as response:  # type: ignore  # noqa: S310
                pte_return = response.read()
        except urllib.error.URLError as error:  # type: ignore
            raise ValueError("No response from device. Check IP address and connections.") from error

        return pte_return



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseBusSchedule class."""

from __future__ import annotations

from bisect import insort
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np

from qililab.constants import PULSEBUSSCHEDULE
from qililab.pulse.pulse_event import PulseEvent
from qililab.utils import Waveforms

if TYPE_CHECKING:
    from qililab.pulse.pulse import Pulse


@dataclass
class PulseBusSchedule:
    """Container of Pulse objects addressed to the same bus.

    Args:
        bus_alias (str): Bus alias.
        timeline (list[PulseEvent]): List of :class:`PulseEvent` objects the PulseBusSchedule is composed of.

    Examples:
        You can create a PulseBusSchedule targeting a bus by doing:

        .. code-block:: python3

            drag_pulse = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            drag_pulse_event = PulseEvent(pulse=drag_pulse, start_time=0)

            drive_schedule = PulseBusSchedule(timeline=[drag_pulse_event], bus_alias="drive_q0")

        You can add further pulse events to an already created PulseBusSchedule. To do so:

        .. code-block:: python3

            drag_pulse_0 = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            drag_pulse_event = PulseEvent(pulse=drag_pulse_0, start_time=0)

            drive_schedule = PulseBusSchedule(timeline=[drag_pulse_event], bus_alias="drive_q0")
            drag_pulse_1 = Pulse(
                amplitude=1, phase=0.5, duration=400, frequency=1e7, pulse_shape=Drag(num_sigmas=2, drag_coefficient=0.2)
            )
            drag_pulse_event_1 = PulseEvent(pulse=drag_pulse_1, start_time=204)
            drive_schedule.add_event(drag_pulse_event_1)
    """

    bus_alias: str  # Bus alias.
    timeline: list[PulseEvent] = field(
        default_factory=list
    )  #: List of PulseEvent objects the PulseBusSchedule is composed of.
    _pulses: set[Pulse] = field(init=False, default_factory=set)

    def __post_init__(self):
        """Sorts timeline and add used pulses to the pulses set if timeline is not empty."""
        if self.timeline:
            self.timeline.sort()
            for pulse_event in self.timeline:
                self._pulses.add(pulse_event.pulse)

    def add_event(self, pulse_event: PulseEvent):
        """Adds pulse event to sequence.

        Args:
            pulse_event (PulseEvent): :class:`PulseEvent` object.
        """
        self._pulses.add(pulse_event.pulse)
        insort(self.timeline, pulse_event)

    @property
    def end_time(self) -> int:
        """End of the PulseBusSchedule.

        Returns:
            int: End of the PulseBusSchedule."""
        end = 0
        for event in self.timeline:
            pulse_end = event.start_time + event.duration
            end = max(pulse_end, end)
        return end

    @property
    def start_time(self) -> int:
        """Start of the PulseBusSchedule.

        Returns:
            int: Start of the PulseBusSchedule."""
        return self.timeline[0].start_time

    @property
    def duration(self) -> int:
        """Duration of the PulseBusSchedule.

        Returns:
            int: Duration of the PulseBusSchedule."""
        return self.end_time - self.start_time

    @property
    def unique_pulses_duration(self) -> int:
        """Duration of the unique pulses, one immediately after the other.

        Returns:
            int: Duration of the unique pulses."""
        return sum(pulse.duration for pulse in self._pulses)

    @property
    def pulses(self):
        """Set of Pulse objects used in this PulseBusSchedule.

        Returns:
            str: The set of :class:`Pulse` objects."""
        return self._pulses

    @property
    def qubit(self):
        """Qubit index addressed by this PulseBusSchedule.

        Raises:
            ValueError: If more than one qubit is addressed by this PulseBusSchedule.

        Returns:
            int: The qubit index addressed by this PulseBusSchedule.
        """
        qubits = {pulse_event.qubit for pulse_event in self.timeline}
        if len(qubits) > 1:
            raise ValueError("More than one qubit is addressed by this PulseBusSchedule.")
        return qubits.pop()

    def __iter__(self):
        """Redirects __iter__ magic method."""
        return self.timeline.__iter__()

    def waveforms(self, resolution: float = 1.0, modulation: bool = True) -> Waveforms:
        """PulseBusSchedule 'waveforms' property.

        Args:
            resolution (float): The resolution of the pulses in ns.

        Returns:
            Waveforms: Class containing the I, Q waveforms for a specific qubit.
        """
        waveforms = Waveforms()
        time = 0
        for pulse_event in self.timeline:
            wait_time = round((pulse_event.start_time - time) / resolution)
            if wait_time > 0:
                waveforms.add(imod=np.zeros(shape=wait_time), qmod=np.zeros(shape=wait_time))
            time += wait_time
            if modulation:
                pulse_waveforms = pulse_event.modulated_waveforms(resolution=resolution)
            else:
                real_env = np.real(pulse_event.envelope(resolution=resolution))
                imag_env = np.imag(pulse_event.envelope(resolution=resolution))
                pulse_waveforms = Waveforms(i=real_env, q=imag_env)
            waveforms += pulse_waveforms
            time += pulse_event.duration

        return waveforms

    def to_dict(self):
        """Returns dictionary representation of the class.

        Returns:
            dict: Dictionary representation of the class.
        """
        return {
            PULSEBUSSCHEDULE.TIMELINE: [pulse_event.to_dict() for pulse_event in self.timeline],
            "bus_alias": self.bus_alias,
        }

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Loads PulseBusSchedule object from dictionary.

        Args:
            dictionary (dict): Dictionary representation of the PulseBusSchedule object.

        Returns:
            PulseBusSchedule: Loaded class.
        """
        timeline = [PulseEvent.from_dict(event) for event in dictionary[PULSEBUSSCHEDULE.TIMELINE]]
        bus_alias = dictionary["bus_alias"]
        return cls(timeline=timeline, bus_alias=bus_alias)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import numpy as np

from qililab.qprogram.blocks import Block
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.waveforms import IQPair, Waveform
from qililab.yaml import yaml


@yaml.register_class
class Calibration:
    """A class to manage calibration data."""

    def __init__(self) -> None:
        """Initialize a Calibration instance."""
        self.waveforms: dict[str, dict[str, Waveform | IQPair]] = {}
        self.weights: dict[str, dict[str, IQPair]] = {}
        self.blocks: dict[str, Block] = {}
        self.crosstalk_matrix: CrosstalkMatrix | None = None

    def add_waveform(self, bus: str, name: str, waveform: Waveform | IQPair):
        """Add a waveform or IQPair for the specified bus.

        Args:
            bus (str): The bus to which the operation will be added.
            operation (str): The name of the operation to add.
            waveform (Waveform | IQPair): The waveform or IQPair instance representing the operation.
        """
        if bus not in self.waveforms:
            self.waveforms[bus] = {}
        self.waveforms[bus][name] = waveform

    def add_weights(self, bus: str, name: str, weights: IQPair):
        """Add a weight for the specified bus.

        Args:
            bus (str): The bus to which the operation will be added.
            operation (str): The name of the operation to add.
            waveform (Waveform | IQPair): The waveform or IQPair instance representing the operation.
        """
        if bus not in self.weights:
            self.weights[bus] = {}
        self.weights[bus][name] = weights

    def add_block(self, name: str, block: Block):
        """Add a block.

        Args:
            name (str): The name of the block to add.
            block (Block): The block to add.
        """
        self.blocks[name] = block

    def has_waveform(self, bus: str, name: str) -> bool:
        """Check if there is an associated waveform for the bus.

        Args:
            bus (str): The bus to check for the waveform.
            operation (str): The name of the waveform to check.

        Returns:
            bool: True if there is an associated waveform with the bus, False otherwise.
        """
        if bus not in self.waveforms or name not in self.waveforms[bus]:
            return False
        return True

    def has_weights(self, bus: str, name: str) -> bool:
        """Check if there are associated weights for the bus.

        Args:
            bus (str): The bus to check for the weights.
            operation (str): The name of the weights to check.

        Returns:
            bool: True if there are associated weights with the bus, False otherwise.
        """
        if bus not in self.weights or name not in self.weights[bus]:
            return False
        return True

    def has_block(self, name: str) -> bool:
        """Check if there is a block with the given name.

        Args:
            name (str): The name of the block to check.

        Returns:
            bool: True if there is a block with the given name, False otherwise.
        """
        return name in self.blocks

    def get_waveform(self, bus: str, name: str):
        """Retrieve waveform for the bus.

        Args:
            bus (str): The bus to check for the waveform.
            name (str): The name of the operation to retrieve.

        Raises:
            KeyError: If the waveform does not exist for the bus.

        Returns:
            Waveform | IQPair: The waveform associated with the bus.
        """
        if bus not in self.waveforms or name not in self.waveforms[bus]:
            raise KeyError(f"The waveform {name} does not exist for the bus {bus}.")
        return self.waveforms[bus][name]

    def get_weights(self, bus: str, name: str):
        """Retrieve weights for the bus.

        Args:
            bus (str): The bus to check for the weights.
            name (str): The name of the weights to retrieve.

        Raises:
            KeyError: If the weights do not exist for the bus.

        Returns:
            IQPair: The weights associated with the bus.
        """
        if bus not in self.weights or name not in self.weights[bus]:
            raise KeyError(f"The weights {name} do not exist for the bus {bus}.")
        return self.weights[bus][name]

    def get_block(self, name: str):
        """Retrieve the block with the given name.

        Args:
            name (str): The name of the block to retrieve.

        Raises:
            KeyError: If the block does not exist.

        Returns:
            Block: The block with the given name.
        """
        if name not in self.blocks:
            raise KeyError(f"The block {name} do not exist.")
        return self.blocks[name]

    def save_to(self, file: str):
        """Dump the current calibration data to a YAML file.

        Args:
            file (str): The file path where the calibration data will be saved. If the file extension
                        is not provided, it defaults to '.yml'.

        Returns:
            str: The path of the saved file.
        """
        yaml.dump(self, Path(file))

        return file

    @classmethod
    def load_from(cls, file: str):
        """Load calibration data from a YAML file.

        Args:
            file (str): The file path from which to load the calibration data.

        Raises:
            TypeError: If the loaded data is not an instance of the Calibration class.

        Returns:
            Calibration: An instance of the Calibration class with data loaded from the file.
        """
        data = yaml.load(Path(file))
        if not isinstance(data, cls):
            raise TypeError("The loaded data is not an instance of the Calibration class.")

        if isinstance(data.crosstalk_matrix, CrosstalkMatrix):
            if isinstance(data.crosstalk_matrix.matrix, list):
                if not data.crosstalk_matrix.bus_list or len(data.crosstalk_matrix.bus_list) != len(
                    data.crosstalk_matrix.matrix
                ):
                    raise ValueError("Bus list is empty or has the wrong dimensions")
                data.crosstalk_matrix.matrix = CrosstalkMatrix.from_array(
                    buses=data.crosstalk_matrix.bus_list, matrix_array=np.array(data.crosstalk_matrix.matrix)
                )

            if isinstance(data.crosstalk_matrix.flux_offsets, list):
                if not data.crosstalk_matrix.bus_list or len(data.crosstalk_matrix.bus_list) != len(
                    data.crosstalk_matrix.flux_offsets
                ):
                    raise ValueError("Bus list is empty or has the wrong dimensions")
                offset_list = data.crosstalk_matrix.flux_offsets.copy()
                data.crosstalk_matrix.flux_offsets = {}
                for bus, value in zip(data.crosstalk_matrix.bus_list, offset_list):
                    data.crosstalk_matrix.flux_offsets[bus] = value
        return data



import copy

import pytest

from qililab.instrument_controllers.yokogawa.gs200_controller import GS200Controller
from qililab.instruments.yokogawa.gs200 import GS200
from qililab.platform import Platform
from qililab.settings import Settings
from tests.data import SauronYokogawa
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronYokogawa.runcard)


@pytest.fixture(name="yokogawa_gs200_controller_wrong_module")
def fixture_yokogawa_gs200_controller_wrong_module(platform: Platform):
    """Return an instance of GS200 controller class"""
    settings = copy.deepcopy(SauronYokogawa.yokogawa_gs200_controller_wrong_module)
    settings.pop("name")
    return GS200Controller(settings=settings, loaded_instruments=platform.instruments)


class TestYokogawaGS200Controller:
    """Unit tests checking the GS200 Controller attributes and methods"""

    def test_initialization(self, platform: Platform):
        """Test QMMController has been initialized correctly"""
        controller_alias = "yokogawa_controller"
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias=controller_alias)
        assert isinstance(controller_instance, GS200Controller)

        controller_settings = controller_instance.settings
        assert isinstance(controller_settings, Settings)

        controller_modules = controller_instance.modules
        assert len(controller_modules) == 1
        assert isinstance(controller_modules[0], GS200)

    def test_check_supported_modules_raises_exception(self, yokogawa_gs200_controller_wrong_module: GS200Controller):
        with pytest.raises(ValueError):
            yokogawa_gs200_controller_wrong_module._check_supported_modules()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QbloxResult class."""
import numpy as np

from qililab.result.qprogram.measurement_result import MeasurementResult
from qililab.typings.enums import ResultName
from qililab.yaml import yaml


@yaml.register_class
class QbloxMeasurementResult(MeasurementResult):
    """QbloxQProgramMeasurementResult class. Contains the acquisitions results for a single measurement obtained from the `Cluster.get_acquisitions` method.

    The input to the constructor should be a dictionary with the following structure:

    - integration: integration data.
        - path_0: input path 0 integration result bin list.
        - path_1: input path 1 integration result bin list.
    - threshold: threshold result bin list.
    - valid: list of valid indications per bin.
    - avg_cnt: list of number of averages per bin.

    Args:
        raw_measurement_data (dict): Raw data obtained from the Qblox digitiser.
    """

    name = ResultName.QBLOX_QPROGRAM_MEASUREMENT

    def __init__(self, bus: str, raw_measurement_data: dict, shape: tuple | None = None):
        super().__init__(bus=bus)
        self.raw_measurement_data = raw_measurement_data
        self.shape = shape

    @property
    def array(self) -> np.ndarray:
        """Get I/Q data as an np.ndarray

        Returns:
            np.ndarray: The I/Q data
        """
        path0 = self.raw_measurement_data["bins"]["integration"]["path0"]
        path1 = self.raw_measurement_data["bins"]["integration"]["path1"]

        array = np.array([path0, path1])
        if self.shape:
            array = array.reshape((2, *self.shape))
        return array

    @property
    def threshold(self) -> np.ndarray:
        """Get the thresholded data as an np.ndarray.

        Returns:
            np.ndarray: The thresholded data.
        """
        array = np.array(self.raw_measurement_data["bins"]["threshold"])
        if self.shape:
            array = array.reshape((1, *self.shape))
        return array
